import os

from scimodels.abstract import APIModel
from scimodels.dependencies import DepedencyLoader

class OpenAIModel(APIModel):
    def __init__(
        self,
        model: str,
        *,
        parallel_count: int | None = 1,
        retries: int | None = 1,
        temperature: float | None = 1.0,
        effort: str | None = None,
        max_tokens: int | None = None
    ):
        self.model = model
        self.temperature = temperature
        self.effort = effort
        self.max_tokens = max_tokens
        self.parallel_count = parallel_count
        self.retries = retries

        # Try to load OpenAI
        dep = DepedencyLoader.load("openai")
        if not dep.loaded:
            raise dep.exception
        self._openai = dep.module

        # Check if OPENAI_API_KEY is set and issue
        # a warning if not
        if not os.getenv("OPENAI_API_KEY"):
            print("SciModels WARNING: your OPENAI_API_KEY enviromental variable is not set. This might cause issues with authentication.")

    def _retrieve_response(
        self,
        query: str
    ) -> str | None:
        request = {
            "model": self.model,
            "messages": [{"role": "user", "content": query}],
            "temperature": self.temperature
        }
        
        if self.effort:
            request["effort"] = self.effort

        if self.max_tokens:
            request["max_tokens"] = self.max_tokens

        response = self._openai.chat.completions.create(**request)

        return response.choices[0].message.content


import os
import json
import string
import random

from scimodels.abstract import BatchModel
from scimodels.temp import get_temp_dir
from scimodels.dependencies import DepedencyLoader

class OpenAIBatchModel(BatchModel):
    def __init__(
        self,
        model: str,
        *,
        parallel_count: int | None = None,
        retries: int | None = None,
        temperature: float | None = 1.0,
        effort: str | None = None,
        max_tokens: int | None = None,
    ):
        self.model = model
        self.temperature = temperature
        self.effort = effort
        self.max_tokens = max_tokens
        
        if parallel_count is not None:
            raise ValueError("'parallel_count' is not supported with this model")
        if retries is not None:
            raise ValueError("'retries' is not supported with this model")
        
        # Try to load OpenAI
        dep = DepedencyLoader.load("openai")
        if not dep.loaded:
            raise dep.exception
        self._openai = dep.module

        # Check if OPENAI_API_KEY is set and issue
        # a warning if not
        if not os.getenv("OPENAI_API_KEY"):
            print("SciModels WARNING: your OPENAI_API_KEY enviromental variable is not set. This might cause issues with authentication.")

    def _send_batch(
        self, 
        queries, 
        *, 
        repeat_count = 1,
    ) -> str:
        temp_dir = get_temp_dir()

        tasks = self._make_prompts(
            queries, 
            self.model, 
            self.temperature,
            self.effort, 
            self.max_tokens, 
            repeat_count
        )
        filename = f"batch-{self._make_id(24)}.jsonl"
        filepath = os.path.join(os.getcwd(), temp_dir, filename)

        os.makedirs(os.path.dirname(filepath), exist_ok=True)

        with open(filepath, 'w', encoding='utf-8') as file:
            for obj in tasks:
                file.write(json.dumps(obj) + '\n')

        batch_file = self._openai.files.create(
            file=open(filepath, 'rb'),
            purpose="batch"
        )

        batch_job = self._openai.batches.create(
            input_file_id=batch_file.id,
            endpoint="/v1/chat/completions",
            completion_window="24h"
        )

        return batch_job.id

    def _fetch(
        self,
        batch_id, 
        *,
        repeat_count: int | None = None,
        stack: bool| None = False,
    ) -> list[str] | list[list[str]] | None:
        
        if repeat_count is not None:
            print("SciModels WARNING: This model automatically handles 'repeat_count'. Your input will be ignored.")

        res = self._openai.batches.retrieve(batch_id)

        if (res.status != "completed"):
            return None

        output_file_id = res.output_file_id
        output_file = self._openai.files.content(output_file_id).text
        output_lines = output_file.splitlines()

        if len(output_lines) == 0:
            return []

        try:
            id_init = json.loads(output_lines[0].strip())['custom_id']
        except:
            return []

        blocks_init = id_init.split('-')
        if len(blocks_init) != 4:
            return []
        
        try:
            total_count = int(blocks_init[3])
            response_count = int(blocks_init[1])
            if stack is None and response_count > 1:
                stack = True
        except:
            return []

        stacked = [[] for _ in range(total_count)]

        for line in output_lines:
            try:
                obj = json.loads(line.strip())

                if 'custom_id' not in obj:
                    continue

                blocks = obj['custom_id'].split('-')

                if len(blocks) != 4:
                    continue
            
                current_index = int(blocks[2])
                result = obj['response']['body']['choices'][0]['message']['content']
                stacked[current_index].append(result)

            except: continue

        # Fill in failed calls with None
        # (or cut if too many calls have been made)
        for lst in stacked:
            count = len(lst)
            if count < response_count:
                lst += [None for _ in range(count, response_count)]
            elif count > response_count:
                lst = lst[:response_count]

        if stack:
            return stacked
        
        unstacked = [
            elem 
            for lst in stacked
            for elem in lst
        ]

        return unstacked

    def _make_id(self, length: int) -> str:
        options = string.ascii_uppercase + string.digits
        chars = [random.choice(options) for _ in range(length)]
        return ''.join(chars)

    def _make_prompts(
        self,
        queries: list[str],
        model: str,
        temperature: float,
        effort: str | None,
        max_tokens: int | None,
        repeat_count: int
    ) -> list[dict]:
        total_count = len(queries)
        return [
            self._make_single_jsonl_prompt(
                query,
                model,
                temperature,
                max_tokens,
                repeat_index,
                repeat_count,
                total_index,
                total_count
            ) for total_index, query in enumerate(queries)
            for repeat_index in range(repeat_count)
        ]

    def _make_single_jsonl_prompt(
        self,
        query: str,
        model: str,
        temperature: float,
        max_tokens: int | None,
        repeat_index: int,
        repeat_count: int,
        total_index: int,
        total_count: int,
    ) -> dict:
        body = {
            "model": model,
            "temperature": temperature,
            "messages": [{
                "role": "user",
                "content": query
            }]
        }

        if max_tokens is not None:
            body["max_tokens"] = max_tokens
        
        return {
            "custom_id": f"{repeat_index}-{repeat_count}-{total_index}-{total_count}",
            "method": "POST",
            "url": "/v1/chat/completions",
            "body": body
        }


from .abstract import Model
from .response import Response
from .loaders import load_model as load
from .loaders import get_providers
from .temp import get_temp_dir, set_temp_dir
from .__version__ import __version__

providers = get_providers()

__all__ = [
    "Model",
    "Response",
    "load",
    "get_temp_dir",
    "set_temp_dir",
    "providers",
]

__license__ = "MIT"

__author__ = "Nikolay Georgiev"


from abc import abstractmethod

from scimodels.response import Response
from scimodels.abstract.modelbase import Model

class BatchModel(Model):
    @property
    def is_batch(self) -> bool:
        return True

    def fetch(
        self,
        batch_id: str,
        *,
        repeat_count: int | None = None,
        stack: bool | None = None,
    ) -> Response:
        super()._ensure_values()
        
        if repeat_count is not None and repeat_count < 1:
            raise ValueError("'repeat_count' must be greater or equal to 1")    
    
        if stack == False and repeat_count is not None and repeat_count > 1:
            raise ValueError("When 'repeat_count' is greater than 1, 'stack' cannot be False")
        
        if stack is None and repeat_count is not None:
            if repeat_count == 1:
                stack = False
            else:
                stack = True
                
        outputs = self._fetch(
            batch_id, 
            repeat_count=repeat_count,
            stack=stack,
        )

        if outputs is None:
            return Response(
                provider=self.provider,
                model=self.model,
                is_ready=False,
                batch_id=batch_id,
                stacked=False,
                outputs=None
            ) 
       
        prestacked = False

        if len(outputs) != 0:
            if isinstance(outputs[0], list):
                stack = True
                prestacked = True

        if stack and not prestacked:
            if repeat_count is None:
                raise ValueError("This model did not automatically handle 'repeat_count', so you must specify it")
            if len(outputs) % repeat_count != 0:
                raise RuntimeError("Invalid 'repeat_count' does not divide number of outputs")
            query_count = len(outputs) / repeat_count
            outputs = [
                outputs[i * repeat_count:(i + 1) * repeat_count]
                for i in range(len(query_count))
            ]

        return Response(
            provider=self.provider,
            model=self.model,
            is_ready=True,
            batch_id=batch_id,
            stacked=(False if stack is None else stack),
            outputs=outputs
        )

    @abstractmethod
    def _fetch(
        self,
        batch_id: str,
        *,
        repeat_count: int | None = None,
        stack: bool | None = None,
    ) -> list[str] | list[list[str]] | None:
        pass

    def _send(
        self,
        queries: list[str],
        *,
        repeat_count: int | None = None,
        stack: bool | None = None,
    ) -> Response:
        if stack is not None:
            raise ValueError("In batch models .send() does not support 'stack'")

        batch_id = self._send_batch(
            queries,
            repeat_count=repeat_count,
        )

        return Response(
            provider=self.provider,
            model=self.model,
            is_ready=False,
            batch_id=batch_id,
            stacked=False,
            outputs=None
        )

    @abstractmethod
    def _send_batch(
        self,
        queries: list[str],
        *,
        repeat_count: int | None = None,
    ) -> str:
        pass


import types
from dataclasses import dataclass

from scimodels.exceptions import DependencyError

@dataclass(frozen=True)
class DependencyResult:
    name: str
    version: str | None = None
    module: types.ModuleType | None = None
    loaded: bool = False
    exception: DependencyError | None = None

    def __post_init__(self):
        if self.loaded:
            if self.module is None:
                raise ValueError("'module' cannot be None when 'loaded' is True")
            if self.exception is not None:
                raise ValueError("'error' must be None when 'loaded' is True")
        else:
            if self.module is not None:
                raise ValueError("'loaded' must the True when 'module' is not None")
            if self.exception is None:
                raise ValueError("'loaded' must be True, when 'error' is None")
            

class DepedencyLoader:
    @staticmethod
    def load(
        dependency_name: str
    ) -> DependencyResult:
        try:
            module = __import__(dependency_name)
            version = module.__version__ if hasattr(module, "__version__") else None
            return DependencyResult(
                name=dependency_name,
                version=version,
                module=module,
                loaded=True,
                exception=None
            )
        except:
            return DependencyResult(
                name=dependency_name,
                version=None,
                module=None,
                loaded=False,
                exception=DependencyError(dependency_name)
            )

    @staticmethod
    def make_exception(install_name: str) -> DependencyError:
        return DependencyError(install_name)


import os
import time

_temp_dir = ".scimodels.tmp/"

def set_temp_dir(dir: str) -> None:
    global _temp_dir
    _temp_dir = dir

def get_temp_dir() -> str:
    if os.path.isabs(_temp_dir):
        return _temp_dir
    current_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(current_dir, _temp_dir)

def clear_temp_dir() -> None:
    temp_dir = get_temp_dir()

    if not os.path.isdir(temp_dir):
        return None
    
    age_limit = 86400 # 1 day
    now = time.time()

    failed_deletes = []

    for filename in os.listdir(temp_dir):
        file_path = os.path.join(temp_dir, filename)

        if os.path.isfile(file_path):
            file_mtime = os.path.getmtime(file_path)
            if now - file_mtime > age_limit:
                try:
                    os.remove(file_path)
                except:
                    failed_deletes.append(filename)
    
    failed_count = len(failed_deletes)

    if failed_count > 0:
        plural = "" if failed_count == 1 else "s"
        file_list = ', '.join(failed_deletes[:4])
        if failed_count > 1:
            more_warning = f" and {failed_count - 4} more"
        else:
            more_warning = ""
        print(f"SciModels WARNING: Failed to delete expired file{plural}: {file_list}{more_warning}.")


from dataclasses import dataclass

@dataclass(frozen=True)
class Response:
    # The LLM/API provider ('openai', 'gemini', etc.)
    provider: str

    # The exact model within the chosen provider
    # ('gtp-5-mini', etc.)
    model: str

    # Indicator if the response is ready
    # (if not - refetch via .refetch())
    is_ready: bool

    # Only for batch models - the fetch Batch ID
    batch_id: str | None

    # If False, all responses are linear in a 1-D list
    # If True, responses are grouped in lists, each list
    # representing a single prompt (useful when each 
    # prompt is sent multiple times via 'repeat' param)
    stacked: bool

    # The final outputs of the call; None when 'is_ready' is
    # False; Otherwise: list[str] when 'stacked' is False; 
    # list[list[str]] when 'stacked' is True 
    outputs: None | list[str] | list[list[str]]

    def __repr__(self) -> str:
        return f"scimodels.Response(provider='{self.provider}', model='{self.model}', is_ready={self.is_ready}, batch_id='{self.batch_id}', stacked={self.stacked}, outputs={self.outputs})"


from abc import abstractmethod
from concurrent.futures import (
    ThreadPoolExecutor, 
    wait, 
    FIRST_COMPLETED,
)
import time

from scimodels.exceptions import MissingParameterError
from scimodels.response import Response
from scimodels.abstract.modelbase import Model

class _Scheduler:
    # Initialize when _Scheduler with a pointer to function 
    # that executes the requests. First argument of that 
    # function must be a prompt string to be send and it 
    # must return a list of responses
    def __init__(
        self, 
        function, 
        concurrent_requests: int | None = 1, 
        max_retries: int | None = 1, 
        delay: float | None = 0.1
    ):
        self.function = function
        self.concurrent_requests = concurrent_requests
        self.max_retries = max_retries
        self.delay = delay
    
    # Run prompts through the function
    def run_queries(
        self, 
        queries: list[str], 
        *args,
        **kwargs
    ):
        total = len(queries)
        completed = 0
        output = [None for _ in range(total)]

        with ThreadPoolExecutor(max_workers=self.concurrent_requests) as executor:
            future_to_index = {
                executor.submit(self.run_with_retries, query, *args, **kwargs): i
                for i, query in enumerate(queries)
            }

            try:
                while future_to_index:
                    done, _ = wait(
                        future_to_index.keys(), timeout=0.2,
                        return_when=FIRST_COMPLETED
                    )
                    for future in done:
                        idx = future_to_index.pop(future)
                        output[idx] = future.result()
                        completed += 1
                        if completed == total:
                            return output
            except KeyboardInterrupt:
                print("Interrupted by user. Shutting down...")
                executor.shutdown(wait=False, cancel_futures=True)
                raise

    # Implement retries
    def run_with_retries(
        self, 
        *args,
        **kwargs
    ) -> list[str] | None:
        try_cnt = 0
        while try_cnt < self.max_retries:
            try:
                output = self.function(*args, **kwargs)
                time.sleep(self.delay)
                if output is not None:
                    return output
                try_cnt += 1
            except Exception as e:
                print(e)
                time.sleep(self.delay)
                # if api error is not due to rate limit, try again
                if "rate limit" not in str(e).lower() and "429" not in str(e):
                    try_cnt += 1
        return None

class APIModel(Model):
    def _check_params(self) -> None:
        if not hasattr(self, "parallel_count"):
            raise MissingParameterError(self, "parallel_count")
        if not hasattr(self, "retries"):
            raise MissingParameterError(self, "retries")

    def _send(
        self,
        queries: list[str],
        *,
        repeat_count: int | None = None,
        stack: bool | None = None,
    ) -> Response:
        self._check_params()

        parallel_count = self.parallel_count
        retries = self.retries

        scheduler = _Scheduler(self._retrieve_response, parallel_count, retries)
        final_queries = [
            query
            for query in queries
            for _ in range(repeat_count)
        ]

        outputs = scheduler.run_queries(final_queries)

        if stack:
            outputs = [
                outputs[i * repeat_count:(i + 1) * repeat_count]
                for i in range(len(queries))
            ]
        
        return Response(
            provider=self.provider,
            model=self.model,
            is_ready=True,
            batch_id=None,
            stacked=stack,
            outputs=outputs
        )
        
    @abstractmethod
    def _retrieve_response(
        self,
        query: str
    ) -> str | None:
        pass


from scimodels.abstract import Model
from scimodels.response import Response
from scimodels.dependencies import DepedencyLoader

class vLLMModel(Model):

    def __init__(
        self,
        model: str,
        *,
        parallel_count: int = 1,
        retries: int | None = None,
        temperature: float = 0.6,
        reasoning_effort: str | None = None,
        max_tokens: int | None = None,
        context_length: int = 2048,
        do_init: bool = True,
    ):
        self.provider = None
        self._model = model
        self._temperature = temperature
        self._effort = reasoning_effort
        self._max_tokens = max_tokens
        self._context_length = context_length
        self._parallel_count = parallel_count
        
        if retries is not None:
            raise ValueError("'retries' is not supported with this model")

        # Try to load vLLM
        vllm = DepedencyLoader.load("vllm")
        if not vllm.loaded:
            raise vllm.exception
        
        # Try to load transformers
        transformers = DepedencyLoader.load("transformers")
        if not transformers.loaded:
            raise transformers.exception

        self._vllm = vllm.module
        self._transformers = transformers

        if do_init:
            self.initialize()

    def initialize(self) -> None:
        LLM = self._vllm.LLM
        SamplingParams = self._vllm.SamplingParams

        self._llm = LLM(
            model=self._model,
            gpu_memory_utilization=0.9,
            max_model_len=self._context_length,
            tensor_parallel_size=self._parallel_count,
        )

        params = {"temperature": self._temperature}

        if self._max_tokens:
            params["max_tokens"] = self._max_tokens
        
        self._tokenizer = self._transformers.AutoTokenizer.from_pretrained(self._model)
        self._params = SamplingParams(**params)

    def _send(
        self,
        queries: list[str],
        *,
        repeat_count: int = 1,
        stack: bool | None = False,
    ) -> Response:
        queries = [
            query for query in queries
            for _ in range(repeat_count)
        ]

        templates = [{
            "role": "user",
            "content": query
        } for query in queries]

        tokenizer_params = {}

        if self._effort:
            tokenizer_params["reasoning_effort"] = self._effort

        prefills = [self._tokenizer.apply_chat_template(
            [query],
            tokenize=False,
            add_generation_prompt=True,
            **tokenizer_params
        ) for query in templates]

        outputs = self._llm.generate(prefills, self._params)
        outputs = [output.outputs[0].text.strip() for output in outputs]

        if stack:
            outputs = [
                outputs[i * repeat_count:(i + 1) * repeat_count]
                for i in range(len(queries))
            ]

        return Response(
            provider=self.provider,
            model=self._model,
            is_ready=True,
            batch_id=None,
            stacked=stack,
            outputs=outputs
        )


from scimodels.abstract import Model

class DependencyError(Exception):
    """Error raised when a Dependency is not installed"""
    def __init__(self, dep_name):
        message = f"""Dependency '{dep_name}' is not installed. Please try running: 
        
        pip install {dep_name}
        """
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.message}"
    
class MissingParameterError(Exception):
    def __init__(self, class_: Model, param_name: str):
        message = f"Class {class_.__class__.__name__} instance missing required parameter '{param_name}'."
        super().__init__(message)
        self.message = message

    def __str__(self):
        return f"{self.__class__.__name__}: {self.message}"


from abc import ABC, abstractmethod

from scimodels.response import Response
from scimodels.temp import clear_temp_dir

class Model(ABC):
    def send(
        self,
        queries: str | list[str],
        *,
        repeat_count: int | None = None,
        stack: bool | None = None
    ) -> Response:
        self._ensure_values()
        clear_temp_dir()

        if repeat_count is None:
            repeat_count = 1

        if repeat_count < 1:
            raise ValueError("'repeat_count' must be greater or equal to 1")    
    
        if stack == False and repeat_count > 1:
            raise ValueError("When 'repeat_count' is greater than 1, 'stack' cannot be False")
        
        if stack is None and not self.is_batch:
            if repeat_count == 1:
                stack = False
            else:
                stack = True

        queries = [queries] if isinstance(queries, str) else queries

        return self._send(
            queries,
            repeat_count=repeat_count,
            stack=stack
        )
    
    @property
    def is_batch(self) -> bool:
        return False

    @abstractmethod
    def _send(
        self,
        queries: list[str],
        *,
        repeat_count: int | None = None,
        stack: bool | None = None
    ) -> Response:
        pass

    def fetch(
        self,
        batch_id: str,
        *,
        repeat_count: int | None = None,
        stack: bool | None = None
    ) -> Response:
        raise RuntimeError("The .fetch() method is only supported by batch models.")
    
    def _ensure_values(self) -> None:
        if not hasattr(self, "provider"):
            self.provider = None
        if not hasattr(self, "model"):
            self.model = None


from scimodels.abstract import Model
from scimodels.models import (
    vLLMModel,
    OpenAIModel,
    OpenAIBatchModel,
)

_SUPPORTED_PROVIDERS = {
    "vllm": vLLMModel,
    "openai": OpenAIModel,
    "openai-batch": OpenAIBatchModel,
}

def load_model(
    provider: str,
    model: str,
    *args,
    **kwargs
) -> Model:
    if provider not in _SUPPORTED_PROVIDERS:
        raise ValueError(f"Provider {provider} is not supported.")
    class_ = _SUPPORTED_PROVIDERS[provider]
    model = class_(model, *args, **kwargs)
    model.provider = provider
    return model

def get_providers() -> list[str]:
    return list(_SUPPORTED_PROVIDERS.keys())